
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <pthread.h>


/* Most of this stuff was pulled from ruby sources as needed */

static pthread_t timer_thread_id = 0;
static pthread_cond_t timer_thread_cond = PTHREAD_COND_INITIALIZER;
static pthread_mutex_t timer_thread_lock = PTHREAD_MUTEX_INITIALIZER;
static volatile int system_working = 1;

#define PER_NANO 1000000000

static struct timespec *
get_ts(struct timespec *ts, unsigned long nsec)
{
    struct timeval tv;
    gettimeofday(&tv, 0);
    ts->tv_sec = tv.tv_sec;
    ts->tv_nsec = tv.tv_usec * 1000 + nsec;
    if (ts->tv_nsec >= PER_NANO) {
	ts->tv_sec++;
	ts->tv_nsec -= PER_NANO;
    }
    return ts;
}

static void rb_disable_interrupt();
static void rb_enable_interrupt();

static void *
thread_timer(void *dummy)
{
    struct timespec ts;

    pthread_mutex_lock(&timer_thread_lock);
    pthread_cond_broadcast(&timer_thread_cond);
#define WAIT_FOR_10MS() pthread_cond_timedwait(&timer_thread_cond, &timer_thread_lock, get_ts(&ts, PER_NANO/100))
    while (system_working > 0) {
	int err = WAIT_FOR_10MS();
	if (err == ETIMEDOUT);
	else if (err == 0 || err == EINTR) {
	    /*if (rb_signal_buff_size() == 0)*/ break;
	}
	/*else rb_bug_errno("thread_timer/timedwait", err);*/

	/*timer_thread_function(dummy);*/
    }
    pthread_mutex_unlock(&timer_thread_lock);
    return NULL;
}

static void rb_thread_reset_timer_thread()
{
    timer_thread_id = 0;
}

static void
rb_thread_start_timer_thread(void)
{
    system_working = 1;
    rb_enable_interrupt();

    if (!timer_thread_id) {
	pthread_attr_t attr;
	int err;

	pthread_attr_init(&attr);
#ifdef PTHREAD_STACK_MIN
	pthread_attr_setstacksize(&attr, PTHREAD_STACK_MIN);
#endif
	pthread_mutex_lock(&timer_thread_lock);
	err = pthread_create(&timer_thread_id, &attr, thread_timer, 0);
	if (err != 0) {
	    pthread_mutex_unlock(&timer_thread_lock);
	    fprintf(stderr, "[FATAL] Failed to create timer thread (errno: %d)\n", err);
	    exit(1);
	}
	pthread_cond_wait(&timer_thread_cond, &timer_thread_lock);
	pthread_mutex_unlock(&timer_thread_lock);
    }
    rb_disable_interrupt(); /* only timer thread recieve signal */
}

static void rb_thread_stop_timer_thread(void)
{
    if (timer_thread_id) {
	int stopped;
	pthread_mutex_lock(&timer_thread_lock);
	stopped = --system_working <= 0;
	if (stopped) {
		pthread_cond_signal(&timer_thread_cond);
	}
	pthread_mutex_unlock(&timer_thread_lock);
	if (stopped) {
		pthread_join(timer_thread_id, 0);
		rb_thread_reset_timer_thread();
	}
    }
}

static void rb_disable_interrupt()
{
    sigset_t mask;
    sigfillset(&mask);
    sigdelset(&mask, SIGVTALRM);
    sigdelset(&mask, SIGSEGV);
    pthread_sigmask(SIG_SETMASK, &mask, NULL);
}

static void rb_enable_interrupt()
{
    sigset_t mask;
    sigemptyset(&mask);
    pthread_sigmask(SIG_SETMASK, &mask, NULL);
}

static void prefork()
{
	rb_enable_interrupt();
	rb_thread_stop_timer_thread();
}

static void postfork()
{
	rb_thread_reset_timer_thread();
	rb_thread_start_timer_thread();
	rb_disable_interrupt();
}


void test_fork()
{
	pid_t pid;

	rb_thread_start_timer_thread();
	usleep(200000);

	printf("Before fork\n");
	fflush(stdout);

	prefork();
	pid = fork();
	postfork();

	printf("After fork pid=%d\n", (int)pid);

	if(pid > 0)
	{
		int st;
		if(waitpid(pid, &st, 0) == -1)
		{
			fprintf(stderr, "waitpid: %s\n", strerror(errno));
		}
		printf("subprocess status %d (0x%x)", st, st);
	}
	exit(0);
}

